// THIS SOURCE CODE IS SUPPLIED "AS IS" WITHOUT WARRANTY OF ANY KIND,
// AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH
// (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL
// WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND ANY
// WARRANTIES OR NON INFRINGEMENT.  THE USER ASSUMES ALL LIABILITY AND
// RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR
// NOR JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE
// FOR DAMAGES OF ANY KIND RESULTING FROM ITS USE.  Without limiting
// the generality of the foregoing, neither the author, nor JAIR, nor
// JAIR's publishers and distributors, warrant that the Source Code
// will be error-free, will operate without interruption, or will meet
// the needs of the user.

#include "solver.h"

// Remove a clause.  Basically just unwatch it, remove it as a reason
// if it is one, and then mark its memory as freed.  You simply refuse
// to erase the bias clause.

void solver::remove_clause(const clause_id ci) 
{
  if (ci == bias_clause) return; // shouldn't be erasing this!
  if (is_reason(ci)) reason[(*ci)[0]] = clause_id();
  unwatch_clause(ci);
  if (!minisat) unset_gamma(ci);
  litpool.release(ci);
}

// Remove all the satisfied clauses.

void solver::remove_satisfied(vector<clause_id> &clauses)
{
  size_t j = 0;
  for (auto cid : clauses)
    if (satisfied(*cid)) remove_clause(cid);
    else clauses[j++] = cid;
  clauses.resize(j);
}

// Remove clauses that don't appear to be worth much.  We do this if
// they're in the bottom half of all the clauses or are worth less
// than the current clause delta divided by the number of clauses.
// Don't remove binary clauses or reasons.  Garbage collect at the
// end.

void solver::reduce()
{
  sort(learned.begin(),learned.end(), [&](clause_id x, clause_id y) {
      return x->size() > 2 && (y->size() == 2 || x->value < y->value); 
    } );
  size_t j = 0;
  for (size_t i = 0; i < learned.size(); ++i) {
    const clause& c = *learned[i];
    if (c.size() > 2 && !is_reason(learned[i]) && 
        (i < learned.size() / 2 || c.value < clause_delta / learned.size()))
      remove_clause(learned[i]);
    else learned[j++] = learned[i];
  }
  learned.resize(j);
  check_garbage();
}

// Simplification, done when you're at the top of the stack and
// intermittently.  Return false if a contradiction is discovered.  If
// you hit the propagation limit, remove any satisfied clauses (you
// can do this because you're at the top of the stack).  Then bump the
// simplificaiton limit by some semi-random amount.

bool solver::simplify()
{
  if (!stack_sizes.empty())             return true;
  if (unit_prop())                      return false; // contradiction found
  if (stack.size() == assignment_limit) return true;
  if (propagations < simplify_limit)    return true;

  remove_satisfied(learned);
  remove_satisfied(clauses);     // is this worthwhile?
  check_garbage();
  reheap();

  assignment_limit = stack.size();
  simplify_limit = propagations + clause_lits + learned_lits;
  return true;
}

// Process a new conflict.  Return false if the theory is unsat and
// true otherwise.  If you're checking nogoods, make sure that the
// learned clause is in fact no good!  Note that we don't use the
// TINISAT trick of unit propagating after the unit clause is learned
// in case we can get another conflict; it seems to be about a push.

// If data is non-NULL, we save the probe prior to the backtrck.  That
// means we record how many choices we've made and variables we've set.

// Once we find the bias conflict, we are always happy to replace it
// with a properly smaller clause (although this actually doesn't seem
// to help much!).  One clause properly subsumes another if it's
// shorter and if every literal in this clause is in the other as
// well.

bool clause::properly_subsumes(const clause &other) const
{
  if (size() >= other.size()) return false;
  for (auto i = begin() ; i != end() ; ++i)
    if (find(other.begin(),other.end(),*i) == other.end()) return false;
  return true;
}

bool clause::properly_subsumes(const clause_id &other) const
{
  return other? properly_subsumes(*other) : true;
}

bool solver::new_conflict(clause_id conf)
{
  ++conflicts;
  if (stack_sizes.empty()) return false; // no backtrack left!
  if (data && conflicts % SAVE_BIAS == 0) save_bias();

  vector<literal> learned_clause = resolve(conf); // resolve and backtrack

  if (!minisat && !bias_clause && learned_clause.size() > 1) {
    vector<literal> bclause = resolve_to_bias(learned_clause);
    if (bclause != learned_clause) bias_clause = new_learned_clause(bclause);
    // save the first backtrack
  }
  if (data) save_probe(learned_clause); // needed to record learned clause length
  backtrack(get_level(learned_clause));

  clause_id why = new_learned_clause(learned_clause);
  make_true(learned_clause[0],why);
  if (!minisat && why->properly_subsumes(bias_clause)) bias_clause = why;
  
  if (learned_clause.size() == 1) {
    known[learned_clause[0]] = true;
    clear_gamma();
  }

  decrease_var_delta();                  // change the increments for the next
  decrease_clause_delta();               // loop

  if (conflicts >= next_increment) {
    learned_limit  *= learned_mult;
    next_increment += display_delta;
    display_delta  *= dd_factor;
    if (verbosity > 0) show_status();
  }
  return true;                           // not known to be unsat
}

void solver::new_clause(const vector<literal> &lits, bool learn)
{
  clause_id cid = litpool.alloc(lits,learn);
  (learn? learned : clauses).push_back(cid);
  watch_clause(cid);
}

clause_id solver::new_learned_clause(vector<literal> &c)
{
  new_clause(c,true);
  if (check_nogoods) assert(model_satisfied(c));
  increase_value(*learned.back());
  return learned.back();
}
