// THIS SOURCE CODE IS SUPPLIED "AS IS" WITHOUT WARRANTY OF ANY KIND,
// AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH
// (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL
// WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND ANY
// WARRANTIES OR NON INFRINGEMENT.  THE USER ASSUMES ALL LIABILITY AND
// RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR
// NOR JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE
// FOR DAMAGES OF ANY KIND RESULTING FROM ITS USE.  Without limiting
// the generality of the foregoing, neither the author, nor JAIR, nor
// JAIR's publishers and distributors, warrant that the Source Code
// will be error-free, will operate without interruption, or will meet
// the needs of the user.

#include <sstream>
#include "solver.h"

// Read a clause.  As long as the input lit is nonzero, get the atom
// and increase the vector of vars and the order_heap as needed.  Also
// increase the vectors used to store all the watchers and Gamma info.

void solver::resize(unsigned n)
{
  unsigned sz = size();
  if (n < sz) return;
  watches.resize(2 * n);
  goccurs.resize(2 * n);
  known  .resize(n,false);
  reason .resize(n);
  level  .resize(n,0);
  value  .resize(n);
  bvalue .resize(n,0);
  model  .resize(n);
  count  .resize(n);
  for (unsigned i = sz ; i < n ; ++i) {
    bias.push_back(literal(i,true));
    polarity.push_back(literal(i,true));
    order_heap.insert_if_missing(i);
  }
}

vector<literal> solver::read_clause(const string &line)
{
  vector<literal> ans;
  int lit;
  for (istringstream is(line) ; is >> lit ; ) {
    if (!lit) break;
    atom a(int_to_lit(lit));
    resize(1 + a);
    ans.push_back(int_to_lit(lit));
  }
  return ans;
}

// Read a problem.  Ignore comments; parse p cnf #vars #clauses line,
// use read_clause to parse anything else.  At the end, check the
// right number of variables and clauses.

bool solver::read(istream &inf)
{
  unsigned nvars = 0, nclauses = 0, cnt = 0;
  for (string line ; getline(inf,line) ; ) {
    if (line.empty() || line[0] == 'c') continue;
    if (line[0] == 'p') {
      assert(line.substr(0,5) == "p cnf");
      istringstream is(line.substr(5));
      is >> nvars >> nclauses;
      continue;
    }
    vector<literal> lits = read_clause(line);
    ++cnt;         // trivial clause won't get added so have to count explicitly
    if (!add_clause(lits)) return false; // unsat detected
  }
  dag.resize(size());
  if (nvars != size()) 
    cerr << "WARNING: wrong number of variables in input" << endl;
  if (nclauses != cnt) 
    cerr << "WARNING: wrong number of clauses in input" << endl;
  return true;
}

// Add a clause, handling some simple clause simplification and
// returning false if a contradiction is found.

// Go through the clause, returning true if there is a known true
// literal or an x/-x pair (they will be adjacent after the sort), and
// dropping any literals known to be false.  Note that we push the
// "interesting" literals onto the original clause, and then resize
// it.

// Now if the clause is empty, it's a contradiction.  If length one,
// make the literal true and try unit propagating.  Otherwise,
// new_clause adds it to the theory.

bool solver::add_clause(vector<literal>& c)
{
  // Check if clause is satisfied and remove false/duplicate literals:
  set<literal> lits(c.begin(),c.end());
  unsigned j = 0;
  for (auto l : lits)
    if (is_true(l)) return true;               // clause includes true lit
    else if (j && l == -c[j - 1]) return true; // clause lincludes l v -l
    else if (not_false(l)) c[j++] = l;
  c.resize(j);

  switch (c.size()) {
  case 0:  return false;
  case 1:  make_true(c[0]);     return !unit_prop();
  default: new_clause(c,false); return true;
  }
}
