// THIS SOURCE CODE IS SUPPLIED "AS IS" WITHOUT WARRANTY OF ANY KIND,
// AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH
// (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL
// WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND ANY
// WARRANTIES OR NON INFRINGEMENT.  THE USER ASSUMES ALL LIABILITY AND
// RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR
// NOR JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE
// FOR DAMAGES OF ANY KIND RESULTING FROM ITS USE.  Without limiting
// the generality of the foregoing, neither the author, nor JAIR, nor
// JAIR's publishers and distributors, warrant that the Source Code
// will be error-free, will operate without interruption, or will meet
// the needs of the user.

#include "solver.h"

// Branch literal.  Just pop off the heap until you find a free
// variable.  If you don't, return the empty literal.

literal solver::pick_branch_pre_backtrack()
{
  atom ans;
  double best = 0;
  for (unsigned a = 0 ; a < size() ; ++a)
    if (is_free(a)) {
      double value = get_uct_value(a);
      if (ans == unsigned(-1) || value > best) { ans = a; best = value; }
    }
  return (ans == unsigned(-1))? literal() : (minisat? polarity[ans] : bias[ans]);
}

literal solver::pick_branch_post_backtrack()
{
  atom next;
  while (!order_heap.empty())
    if (is_free(next = order_heap.front())) return polarity[next];
  return literal();
}

literal solver::pick_branch(bool pre_backtrack)
{
  ++decisions;
  return (pre_backtrack & !minisat)? pick_branch_pre_backtrack() : 
    pick_branch_post_backtrack();
}

// Increase the value of a variable or a clause.  Simple -- just bump
// it, and if it's too big, scale everything back.  Also update the
// biggest heuristic value, which is used to normalize things if we're
// recording performance.

void solver::increase_value(atom a) 
{
  static double vmax = 1e100;
  bvalue[a] += variable_delta;
  heuristic_max = max(heuristic_max,bvalue[a]);
  if (bvalue[a] > vmax ) {
    for (size_t i = 0 ; i < size() ; ++i) bvalue[i] *= 1 / vmax;
    variable_delta *= 1 / vmax; 
    heuristic_max *= 1 / vmax;
  }
  order_heap.update(a); 
}

void solver::increase_value(clause& c) 
{
  static float cmax = 1e20;
  if ((c.value += clause_delta) > cmax ) {
    for (auto &i : learned) i->value *= 1 / cmax;
    clause_delta *= 1 / cmax; 
  }
}

// For some reason that I don't understand, it's significantly better
// to rebuild the heap this way than to insert all the elements onto
// it (keeping the smallest at the front as you go).  It's not just
// timing; the number of nodes expanded is significantly smaller as
// well!

void solver::reheap()
{
  vector<atom> v;
  for (size_t i = 0 ; i < size() ; i++) if (is_free(i)) v.push_back(i);
  order_heap.rebuild(v);
}

void solver::update_variable_counts()
{
  for (auto ss : stack_sizes) ++count[stack[ss]];
}

// UCT primitive.  Just like paper -- val is estimated value, ct is
// the number of times we've tried this variable, pulls is total
// number of attempts, c is tuning factor.

double uct_value(double val, double ct, size_t pulls, double c)
{
  return val + c * sqrt(2 * pulls / max(0.01,ct));
}

// Here is the value of a variable according to UCT.  It's basically
// the VSIDS value, which might or might not match the polarity.

double solver::get_uct_value(const atom a) const
{
  double raw = bvalue[a];
  if (bias[a] != polarity[a]) raw = -raw;
  return uct_value(raw,count[a],starts,explore);
}

// Here is the value of a particular choice of literal biases in the
// case where you're possibly flipping multiple literals.  It's
// basically the value you get from all the literals that you actually
// flip, where assigned_val computes the new bias value of a literal
// and equiv contains the current value.  The sign is negated because
// get_uct_value returns a *positive* value if you *want* to set the
// variable as indicated by the bias.

double solver::get_uct_value(int chosen, const vector<atom> &equiv) const
{
  double ans = 0;
  for (size_t i = 0 ; i < equiv.size() ; ++i)
    if (equiv[i].assigned_val(i,chosen) != value[equiv[i]])
      ans -= get_uct_value(equiv[i]);
  return ans;
}

// Value of a literal, either negated or not depending on whether or
// not the variable is in vals.

literal atom::assigned_val(size_t idx, int vals) const
{
  bool bit = vals & (1 << idx);
  return literal(*this,!bit);   // bit false means negated
}

// Here we set the bias to the best possibility given the above
// values.  _possibles_ is a vector of binary possibilities for the
// _equiv_ group of atoms, where a binary value is possible if the
// collected nogoods haven't eliminated it.  The first piece of code
// finds the best choice, and the second sets the bias appropriately.

void solver::set_best_possibility(const vector<int> &possibles, 
                                  const vector<atom> &equiv)
{
  double best = 0;
  int chosen = 0;
  for (auto p : possibles) {
    double val = get_uct_value(p,equiv);
    if (p == possibles[0] || val > best) { chosen = p; best = val; }
  }

  for (size_t i = 0 ; i < equiv.size() ; ++i)
    set_bias(equiv[i].assigned_val(i,chosen));
}

// A culprit for a particular clause is *disallowed* if it's a bad
// conclusion (unit and unflippable, basically).  We also only return
// the first element of an equivalence class as a possible conclusion;
// there is no need to return the others.

bool solver::ok_conclusion(literal l, const clause &c) const
{
  if (bad_conclusion(l)) return false;
  for (size_t i = 0 ; c[i] != l ; ++i)
    if (!bad_conclusion(c[i]) && equiv(c[i],l)) return false;
  return !over_width(l,c);
}

// Will using a particular literal l as the conclusion result in an
// oversize equivalence class?  We have to find all the atoms a that
// both follow l and precede an element of c.  That's basically easy,
// but we want to exploit the fact that the preceders and follower
// lists are sorted by atom number.  So as we search, we keep track of
// where we "are" in the preceders of each element of c; that narrows
// the set of things we have to look at as we search for each follower
// of l.  And if we're over the limit at any point, we can stop
// looking.

bool solver::over_width(literal l, const clause &c) const
{
  vector<const VarSet *> all_preceders;
  vector<atom> atoms;
  for (auto lit : c)
    if (!known[lit]) all_preceders.push_back(&dag.preceders(atom(lit)));
  if (all_preceders.empty()) return false;
  size_t lim = 1 + width_limit();
  vector<size_t> ptrs(all_preceders.size(),0);
  for (auto a : dag.followers(atom(l))) 
    for (size_t j = 0 ; j < all_preceders.size() ; ++j) {
      while (ptrs[j] < all_preceders[j]->size() && 
             (*all_preceders[j])[ptrs[j]] < a) ++ptrs[j];
      if (ptrs[j] != all_preceders[j]->size() && 
          (*all_preceders[j])[ptrs[j]] == a) { 
        ++ptrs[j]; 
        if (--lim == 0) return true; // found it so ignore other preceder sets
        break;
      }
    }
  return false;
}

// Here's the function that actually does it all, collecting the
// possibilities for set_best_possibility.  Note that we just pick the
// single best literal, not trying to construct the equivalent classes
// and evaluate them until later.

literal solver::find_conclusion(const clause &c) 
{
  literal ans = 0;
  double best = 0;
  for (auto lit : c)
    if (ok_conclusion(lit,c)) {
      double value = get_uct_value(atom(lit));
      if (ans == 0 || value > best) { ans = lit; best = value; }
    }
  return ans;
}
