// THIS SOURCE CODE IS SUPPLIED "AS IS" WITHOUT WARRANTY OF ANY KIND,
// AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH
// (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL
// WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND ANY
// WARRANTIES OR NON INFRINGEMENT.  THE USER ASSUMES ALL LIABILITY AND
// RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR
// NOR JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE
// FOR DAMAGES OF ANY KIND RESULTING FROM ITS USE.  Without limiting
// the generality of the foregoing, neither the author, nor JAIR, nor
// JAIR's publishers and distributors, warrant that the Source Code
// will be error-free, will operate without interruption, or will meet
// the needs of the user.

#include <iomanip>
#include <fstream>
#include <csignal>
#include <unistd.h>
#include <cstdlib>
#include <cstring>
#include <signal.h>

using namespace std;

#include "solver.h"

static string FLEX_VERSION = "0.1";
pool *clause_id::p = nullptr;

string print_options()
{
  return "flexsat " + FLEX_VERSION + " options:\n"
    "\n"
    " -b \tBacktrack compulsively\n"
    " -c n\tClause decay factor (default 0.999)\n"
    " -C n\tCPU limit in sec\n"
    " -d \tDebug\n"
    " -f n\tFirst restart interval (default 100)\n"
    " -g n\tMemory fraction freed before Garbage collection (default 0.2)\n"
    " -l n\tprobe Limit\n"
    " -m \tshow Model\n"
    " -M n\tMemory limit in MB\n"
    " -n \tcheck Nogoods\n"
    " -p name\tmeasure Performance, dump output in name\n"
    " -P name\tmeasure Performance, append output to name\n"
    " -s \tSimulate minisat\n"
    " -t \tTotal order\n"
    " -T n\tTune Monte Carlo parameter\n"
    " -v n\tVariable decay factor (default 0.95)\n"
    " -V n\tVerbosity (0-2, default 1)\n"
    " -w n\tWidth limit (linear)\n"
    " -W n\tWidth limit (log)\n"
    " -z \tInput file is gZipped\n"
    "\n"
    "Example:\n"
    "\t./stlsat problem.cnf -lM 100\n"
    "\n";
}

string print_usage()
{
  return "Usage: flexsat [cnf-file-name] [options]\n"
    "Solve the SAT problem specified in cnf-file-name (or stdin).\n"
    "Example: stlsat sat-problem.cnf\n"
    "\n" + print_options();
}

float get_int(size_t num, char** argv, size_t &idx, char c)
{
  if (++idx >= num) {
    cerr << "Not enough arguments for " << c << endl;
    exit(1);
  }
  return atoi(argv[idx]);
}

float get_float(size_t num, char** argv, size_t &idx, char c)
{
  if (++idx >= num) {
    cerr << "not enough arguments for " << c << endl;
    exit(1);
  }
  return atof(argv[idx]);
}

string get_string(size_t num, char** argv, size_t &idx, char c)
{
  if (++idx >= num) {
    cerr << "not enough arguments for " << c << endl;
    exit(1);
  }
  return argv[idx];
}

settings::settings(size_t num, char** argv)
{
  *this = settings();
  for (size_t i = 1 ; i < num ; ++i)
    if (argv[i][0] == '-') {
      size_t j = i;
      for (size_t k = 1 ; argv[j][k] ; ++k) {
        switch (argv[j][k]) {
        case 'b': compulsive = true; restart_first = 1;         break;
        case 'c': clause_decay = get_float(num,argv,i,'c');     break;
        case 'C': cpu_lim = get_int(num,argv,i,'C');            break;
        case 'd': debug = true;                                 break;
        case 'f': restart_first = get_int(num,argv,i,'f');      break;
        case 'g': garbage_frac = atof(argv[++i]);               break;
        case 'l': probe_lim = get_int(num,argv,i,'l');          break;
        case 'm': print_model = true;                           break;
        case 'M': mem_lim = get_int(num,argv,i,'M');            break;
        case 'n': check_nogoods = true;                         break;
        case 'p': performance = true;                           break;
        case 's': minisat = true;                               break;
        case 't': width = 1; log_width = 0;                     break;
        case 'T': explore = get_float(num,argv,i,'T');          break;
        case 'v': var_decay = get_float(num,argv,i,'v');        break;
        case 'V': verbosity = get_int(num,argv,i,'V');          break;
        case 'w': width = get_float(num,argv,i,'w');            break;
        case 'W': log_width = get_float(num,argv,i,'W');;       break;
        case 'x': fastexit = true;                              break;
        case 'z': gzip = true;                                  break;
        default: cerr << print_usage(); exit(1);
        }
      }
    }
    else if (!infile.empty()) {
      cerr << "Too many files specified\n";
      exit(1);
    }
    else infile = argv[i];
}

void settings::set_cpu_limit() const
{
  if (cpu_lim != size_t(-1)) set_limit(RLIMIT_CPU,cpu_lim,"CPU time");
}

void settings::set_memory_limit() const
{
  if (mem_lim != size_t(-1))
    set_limit(RLIMIT_AS,rlim_t(mem_lim * 1024 * 1024), "memory");
}

// Show normal summary stuff?  Yes, if verbosity is positive and
// you're not showing performance details.

bool solver::show_summary() const
{
  if (verbosity <= 0) return false;
  if (verbosity > 1 && data != nullptr) return false;
  return true;
}

void solver::show_header(double t) const
{
  if (!show_summary()) return;
  unsigned p = cout.precision();
  cout << "=======================[ Problem Statistics ]"
    "========================\n"
       << "|                                                 "
    "                  |\n"
       << "|  Number of variables:  " << setw(12) << size()
       << "                               |\n"
       << "|  Number of clauses:    " << setw(12) << clauses.size()
       << "                               |\n"
       << "|  Parse time:           " << setw(12) << fixed << setprecision(2) 
       << t << setprecision(p) << " s                             |\n"
       << "|                                                 "
    "                  |\n"
       << "=======================[ Search Statistics ]======"
    "===================\n"
       << "| Conflicts |          ORIGINAL         |         "
    " LEARNED          |\n"
       << "|           |    vars  clauses     lits |    limit"
    "  clauses lits/cl |\n"
       << "=================================================="
    "===================\n";
}

void solver::show_trailer() const
{
  if (show_summary())
    cout << "=================================================="
      "===================\n";
}

void solver::show_unit() const
{
  if (show_summary())
    cout << "================================================"
      "=====================\n"
         << "Solved by unit propagation\n" << statistics(*this) << endl;
  cout << "UNSATISFIABLE\n";
  exit(20);
}

void solver::show_model() const
{
  cout << "model";
  for (size_t i = 0; i < size(); i++) 
    if (!model[i].negated()) cout << ' ' << model[i];
  cout << " 0" << endl;
}

void solver::show_result(Result r) const
{
  if (show_summary()) cout << statistics(*this) << endl;
  switch (r) {
  case SAT:
    if (verbosity > 1 || print_model) show_model();
    cout << "SATISFIABLE\n";   
    break;
  case UNSAT:   cout << "UNSATISFIABLE\n"; break;
  case UNKNOWN: cout << "INDETERMINATE\n"; break;
  }
}

void show_failure()
{
  cout << "=================================================="
    "===================\n"
       << "INDETERMINATE";
  cout << endl;
  exit(0);
}

// Read a zipped file.  Unzip it and read it.  We do it this way to
// avoid having to recompile if a gzip vulnerability is discovered.

bool solver::read_zip(const string &inf)
{
  char f[L_tmpnam];
  strcpy(f,"/tmp/fileXXXXXX");
  mktemp(f);
  system(("gunzip -c " + inf + " > " + string(f)).c_str());
  ifstream ifs(f);
  bool ans = read(ifs);
  remove(f);
  return ans;
}

// Read in theory; return false if known unsat (empty clause, etc).
// If there is a file to dump performance data to, create a new
// problem_data to store it.  Doing that involves just grabbing the
// number of variables and initial number of clauses when the problem
// is set up.

problem_data::problem_data(const solver &slv) : 
  nvars(slv.size()), initial_clauses(slv.clauses.size()) { }

bool solver::initialize()
{
  initial_time = get_cpu_time();
  clause_id::p = &litpool;

  set_cpu_limit();
  set_memory_limit();

  bool ans;
  if (infile.empty()) ans = read(cin);
  else if (gzip) ans = read_zip(infile);
  else {
    ifstream inf(infile);
    ans = read(inf);
  }

  if (performance) {
    data = new problem_data(*this);
    save_bias();
    if (verbosity > 1) performance = ""; // so you output to cout
  }
  if (show_summary()) show_header(get_cpu_time() - initial_time);
  return ans;
}

Result solver::solve(bool init_ok)
{
  Result ans;
  if (!init_ok) ans = UNSAT;
  else if (!simplify()) { show_unit(); ans = UNSAT; }
  else ans = solve();
  if (verbosity >= 0) show_result(ans);
  return ans;
}

void solver::copy_model(const solver &other)
{
  for (size_t i = 0 ; i < size() ; ++i) model[i] = other.model[i];
}

// Force quit.  Show the summary if requested; call _exit to make sure
// nothing untoward can happen.

void solver::force_exit()
{
  cout << endl << "*** ABORT ***" << endl;
  if (show_summary())
    cout << statistics(*this) << endl << "*** ABORT ***" << endl;
  _exit(1);
}

// Record an interrupt.

void solver::interrupt()
{
  interrupted = true;
}

// Main solver.  Only catch is that if you're checking the nogoods,
// first solve via minisat-equivalent, and then check that the nogoods
// you derive are in fact no good!

int main(int argc, char** argv)
{
  static solver *slv;           // needed for interrupt handler
  try {
    settings sets(argc,argv);
    solver s(sets);
    slv = &s;
    signal(SIGINT, [](int sig) { slv->force_exit(); });
    signal(SIGXCPU,[](int sig) { slv->force_exit(); });
    bool ok = s.initialize();
    if (s.check_nogoods) {
      solver s2(sets);
      s2.check_nogoods = false;
      s2.minisat = true;
      s2.verbosity = -1;
      Result r = s2.solve(s2.initialize());
      if (r != SAT) {
        cerr << "Can't check nogoods for " << ((r == UNSAT)? "UNSAT" : "UNKNOWN")
             << " theory!" << endl;
        return 1;
      }
      s.copy_model(s2);
    }
    signal(SIGINT, [](int sig) { slv->interrupt(); });
    signal(SIGXCPU,[](int sig) { slv->interrupt(); });
    Result ans = s.solve(ok);

    int rval = (ans == SAT ? 10 : ans == UNSAT ? 20 : 0);
    if (s.fastexit) exit(rval);
    return rval;
  } catch (...) { show_failure(); }
}
