// THIS SOURCE CODE IS SUPPLIED "AS IS" WITHOUT WARRANTY OF ANY KIND,
// AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH
// (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL
// WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND ANY
// WARRANTIES OR NON INFRINGEMENT.  THE USER ASSUMES ALL LIABILITY AND
// RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR
// NOR JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE
// FOR DAMAGES OF ANY KIND RESULTING FROM ITS USE.  Without limiting
// the generality of the foregoing, neither the author, nor JAIR, nor
// JAIR's publishers and distributors, warrant that the Source Code
// will be error-free, will operate without interruption, or will meet
// the needs of the user.

#include "solver.h"

// Mark a clause as belonging to gamma.

void solver::set_gamma(clause_id c, literal conclusion)
{
  const clause &cls = *c;
  if (cls.gamma) return;
  cls.gamma = true;
  gamma.insert(c);
  for (auto l : cls) goccur(l).push_back(c);
  cls.conclusion = conclusion;
}

// Remove a clause from gamma.

static inline void remove(vector<clause_id>& cs, const clause_id& t)
{
  auto i = find(cs.begin(),cs.end(),t);
  *i = cs.back();
  cs.pop_back();
}

void solver::unset_gamma(const clause_id c)
{
  const clause &cls = *c;
  if (!cls.gamma) return;
  cls.gamma = false;
  gamma.erase(c);
  for (auto l : cls) remove(goccur(l),c);
}

// Clear Gamma completely.  Unmark all clauses, clear Gamma set, clear
// gamma_arr entries.  Clear the dag, which we can do with resize.

void solver::clear_gamma()
{
  for (auto &c : gamma) c->gamma = false;
  gamma.clear();
  for (auto &g : goccurs) g.clear();
  dag.resize(size());
}

// PDB analysis to fold in new nogood; return false if unsat proven.
// Basically Procedure 5.7 of newsys.pdf, lines 10-15.  Lines 10-11
// are here and 12-13 are in adjust_bias, below.  Lines 14-15 are
// handled by the overall search loop, which will backtrack to the
// root at this point.

bool solver::absorb_nogood()
{
  clause_id c = simp(bias_clause);              // line 10
  if (c == clause_id()) return false;           // line 11
  if (c->size() > 1) adjust_bias(c);            // otherwise, clear bias anyway
  else simplify();
  return true;
}

// Lines 12-13 of Procedure 5.7 of newsys.pdf.  gamma_conflict sets
// all possible values in possibles and puts the atoms in the
// equivalence class in equiv.  set_best_possibility picks the best
// result and does what it says.

void solver::adjust_bias(const clause_id cl)
{
  literal l = cl->conclusion;
  vector<int> possibles;
  vector<atom> equiv;
  gamma_conflict(l,&possibles,&equiv);
  set_best_possibility(possibles,equiv); // in heuristic.cpp
}

// The hard part above is gamma_conflict, which is supposed to figure
// out the set of possible assignments for literals equivalent to l.
// Finding the literals equivalent to l is easy; figuring out what's
// allowed is harder.

// The lowest level piece of this involves testing a single clause for
// satisfaction given some setting for the variables in the
// equivalence class, ant this function does that.  vals is the int in
// [0,2^n), indices gives the index in [0,n) of each atom in the whole
// theory (-1 if it's missing), and cls is the clause itself.
// Literals in the clause that are not mentioned in indices are
// assumed to be false, since the clause itself is unsatisfied.

bool solver::satisfy(int vals, const vector<int> &indices, const clause &cls)
  const
{
  for (auto l : cls) {
    atom a(l);
    if (indices[a] >= 0 && l == a.assigned_val(indices[a],vals)) return true;
  }
  return false;
}

// Given indices for each variable in the theory, does a particular
// element of [0,2^n) break at least one of the associated clauses in
// Gamma?  vals is the element as usual, clauses is the clauses you
// care about (not all of gamma; just the ones where the relevant
// variable appears), eq is the equivalence class of the
// variable that started it all, and indices is the index of each
// variable in the theory.

// This is pretty easy; we check the clauses singly.  For each clause
// c, if the conclusion is in the equivalence class and it's not
// satisfied, that's a failure.  If there aren't any failures, then
// the gamma set is ok.

bool solver::breaks_gamma(int vals, const vector<clause_id> &clauses,
                          const VarSet &eq, const vector<int> &indices) const
{
  for (auto &cid : clauses)
    if (eq.includes(atom(cid->conclusion)) && !satisfy(vals,indices,*cid)) 
      return true;
  return false;
}

// Now we put it all together, which is mostly a matter of setting up
// an appropriate call to breaks_gamma above.  First, we get the
// equivalence class of the literal in question.  We reuse the vector
// of indices, since it will mostly be -1 at any given point and
// there's no reason to reconstruct it.

// In building up the indices, we can ignore any variable in the
// equivalence class that's unit; those will have already been
// handled.  So for the non-unit variables, we set the index
// appropriately; we keep track of all the indices because we want to
// *unset* the entries in indices so that we can reuse that static
// vector.  We also collect in sclauses all the clauses where a
// relevant literal appears; that's what breaks_gamma expects.

// Having done all the work, we call breaks_gamma for each i in
// [0,2^n), accumulating all of the values for which gamma is ok.  As
// soon as we find one such value, we know that we're ok from a
// conflict perspective.  At the end, we reset indices as described in
// the previous paragraph.

// There is one further subtlety.  If possibles is non-NULL, we want
// to collect the possible values for the atoms in the equivalence
// class.  But it's tricky because we preferentially want to select
// values where l is true (after all, l is the designated conclusion
// of some new nogood, so presumably we want to flip it).  However,
// there might not be any!  So we push all the values for which l is
// true onto possibles; then if possibles winds up empty, we copy the
// backups over instead.

bool solver::gamma_conflict(literal l, vector<int> *possibles, 
                            vector<atom> *new_indices) const
{
  const VarSet &eq = dag.equiv_class(atom(l));
  static vector<int> indices;
  if (indices.size() < size()) indices.resize(size(),-1);
  vector<atom> tmp_indices;
  if (!new_indices) new_indices = &tmp_indices;
  set<clause_id> sclauses;        // clauses where any relevant literal appears
  for (auto e : eq) {
    if (!is_free(e) && known[e]) continue; // don't care about units
    indices[e] = new_indices->size();
    new_indices->push_back(e);
    sclauses.insert(goccur(e).begin(),goccur(e).end());
    sclauses.insert(goccur(-e).begin(),goccur(-e).end());
  }
  vector<clause_id> clauses(sclauses.begin(),sclauses.end());
  bool fail = true;
  int lindex = indices[(atom) l];
  vector<int> backups;

  for (int i = 0 ; i < (1 << new_indices->size()) ; ++i)
    if (!breaks_gamma(i,clauses,eq,indices)) {
      fail = false;
      if (possibles) 
        if (((atom) l).assigned_val(lindex,i) == l) possibles->push_back(i);
        else backups.push_back(i);
      else break;
    }
  if (possibles && possibles->empty()) *possibles = backups;
  for (auto n : *new_indices) indices[n] = -1;
  return fail;
}

// Procedure 5.4 of newsys.pdf, with some simple ordering mods.  (1)
// We add c to Gamma at the beginning so resolution and stuff is
// easier.  (2) We do lines 5-7 first; if we're going to resolve, we
// don't need to adjust the partial order before doing so.

clause_id solver::simp(clause_id cid)
{
  literal l = find_conclusion(*cid);               // line 2
  if (is_false(l) && known[l]) return clause_id(); // hopeless (can this happen?)
  if (cid->size() > 1) set_gamma(cid,l);           // line 9
  if (gamma_conflict(l)) {                         // line 5
    unset_gamma(cid);                              // undo set from above
    vector<literal> c = pdb_resolve(l,cid);        // line 6
    if (c.empty()) return clause_id();             // hopeless
    find_level(c);
    clause_id tmp = new_learned_clause(c);
    return simp(tmp);                              // line 7
  }

  atom a(l);
  for (auto lit : *cid) if (!known[lit]) add_arc(atom(lit),a);  // line 3
  weaken(a);                                                    // line 4
  cid->conclusion = l;
  if (cid->size() == 1) clear_gamma();
  else simplify_gamma(l);                                       // line 8

  return cid;
}

// The final block of code deals with simplifying Gamma in line 8.
// Here is the first part of that, removing from Gamma any nogood with
// l in its antecedent.  The various possible l are identified in the
// main call to simplify_gamma below.

void solver::simplify_antecedent(literal l)
{
  vector<clause_id> unsetme;
  for (auto &cid : goccur(l))
    if (!equiv(l,cid->conclusion)) unsetme.push_back(cid);
    // can't unset inline because goccurs gets messed up!
  unset_gamma(unsetme);
}

// The other clause in line 8 is a bit more intricate.  Here we have a
// specific literal l (as in the description in the paper) and want to
// remove clauses containing l *as conclusion* with disjunctive
// conclusions.  Note that we call this function for *every* literal
// that follows x in the partial order, so we will eventually call it
// with its conclusion, whatever that conclusion may be.

void solver::make_parseable_literal(literal l)
{
  vector<clause_id> unsetme;
  for (auto &cid : goccur(l)) {
    if (l == cid->conclusion) continue;
    bool found = false;
    for (auto lit : *cid)
      if (leq(lit,l) == leq(l,lit)) { // both true (equiv) or both false (incomp)
        if (found) { unsetme.push_back(cid); break; }
        else found = true;
      }
  }
  unset_gamma(unsetme);
}

// Same thing as above, but instead of having a literal, we now have
// an atom.  So we do it for the literals equivalent to that atom.

void solver::make_parseable(atom a)
{
  for (auto eq : dag.equiv_class(a)) {
    make_parseable_literal(eq);
    make_parseable_literal(-eq);
  }
}

// And here we assemble the pieces.  We have to call the above
// function for every atom that follows the given one.

void solver::simplify_gamma(literal l)
{
  for (auto eq : dag.equiv_class(atom(l))) {
    simplify_antecedent(eq);
    simplify_antecedent(-eq);
  }
  for (auto pr : dag.proper_followers(atom(l))) make_parseable(pr);
}

// Here's the basic resolvent procedure (line 6 of Procedure 5.4 of
// newsys.pdf).  Basically, we begin by getting the antecedent of the
// known conflict clause, and then have to merge in the result of
// calling the function below with each literal that's equivalent to
// the fixed conclusion.  Then we combine all the literals into a
// clause.

vector<literal> solver::pdb_resolve(literal l, const clause_id cid)
{
  set<literal> resolvent;
  for (auto lit : *cid) if (!equiv(lit,l)) resolvent.insert(lit);
  set<clause_id> done;
  for (auto eq : dag.equiv_class(atom(l))) {
    pdb_resolve(eq,done,resolvent);
    pdb_resolve(-eq,done,resolvent);
  }
  return vector<literal>(resolvent.begin(),resolvent.end());
}

// Resolution step (line 6) of simp procedure.  Handle each clause
// once, of course.  For each clause including a specific literal, if
// the conclusion is equivalent to the literal, push everything in the
// clause that *isn't* equivalent to that literal into the resolvent.

void solver::pdb_resolve(literal l, set<clause_id> &done, 
                         set<literal> &resolvent) const
{
  for (auto &cid : goccur(l)) {
    if (!equiv(l,cid->conclusion)) continue;    // not in conclusion
    if (done.find(cid) != done.end()) continue; // already did it
    done.insert(cid);
    for (auto lit : *cid) if (!equiv(l,lit)) resolvent.insert(lit);
  }
}
