// THIS SOURCE CODE IS SUPPLIED "AS IS" WITHOUT WARRANTY OF ANY KIND,
// AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH
// (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL
// WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND ANY
// WARRANTIES OR NON INFRINGEMENT.  THE USER ASSUMES ALL LIABILITY AND
// RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR
// NOR JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE
// FOR DAMAGES OF ANY KIND RESULTING FROM ITS USE.  Without limiting
// the generality of the foregoing, neither the author, nor JAIR, nor
// JAIR's publishers and distributors, warrant that the Source Code
// will be error-free, will operate without interruption, or will meet
// the needs of the user.

#include "solver.h"

// Debugging stub

void solver::check(const string &desc) const 
{
}

// PRIVATE METHODS

// Setting the bias or polarity.  Only tricky bit is that if you're
// recording all the data and it changed, then increment the count
// associated with this literal.  Do that for the bias in the FLEX
// case and the polarity in the MINISAT case.

void solver::set_bias(literal l)
{
  if (data && !minisat && bias[l] != l) increment_heuristics(l);
  bias[l] = l; 
}

void solver::set_polarity(literal l)
{
  if (data && minisat && polarity[l] != l) increment_heuristics(l);
  polarity[l] = l; 
}

// Assign a value to a variable.  Set the value, reason, and level.
// Add the literal the current assignment.

void solver::make_true(literal p, const clause_id why)
{
  assert(is_free((atom) p));
  value[p]  = p;
  reason[p] = why;
  level[p]  = curr_level();
  if (curr_level() == 0) set_bias(p);
  stack.push_back(p);
}

// Backtrack.  Easy -- just free the variables, and save the current
// choice as the intended polarity of the given variable.  Push the
// variable back onto the heap as well.  In the loop, note that we
// have to be careful to handle c=-1 correctly (the comparison would
// default to unsigned).

void solver::backtrack(unsigned level) 
{
  if (curr_level() > level) {
    for (int c = stack.size() - 1 ; c >= int(stack_sizes[level]) ; --c) {
      atom a(stack[c]);
      value[a]    = literal();
      set_polarity(stack[c]);
      order_heap.insert_if_missing(a); 
    }
    stack.resize(propagation_point = stack_sizes[level]);
    stack_sizes.resize(level);
  }
}

// Resolution step.  conf is the clause that is essentially
// contradicted by the current assignment.  We want to return the new
// clause.

vector<literal> solver::resolve(clause_id conf) 
{
  static fvector<char> inclause; // lits in clause
  if (inclause.empty()) inclause.resize(size(),false);
  vector<literal> ans = combine_reasons(conf,inclause); // produces a UIP
  ans = simplify(ans,inclause);
  find_level(ans);          // first lit is asserted; second is backtrack level
  return ans;
}

// Resolution starts with a conflict clause, and then constructs an
// FUIP from that.  So for a start, we have to resolve away literals
// at the assertion level until there is just one such literal
// remaining (i.e., construct a UIP).  That's pretty easy.  The hard
// part is that we then have to remove from the resulting clause any
// literals that can be resolved away without introducing any new
// literals at all (which is handled by simplify, below)

// We construct the UIP _ans_, which has all the literals in *conf at
// any level other than the current one and knows when to stop by
// counting the number of literals at the current level.

// We work back through the stack until we find an inclause literal
// (one that was marked as part of *conf).  We get the reason for this
// literal and will add the literals in *that* clause on the next pass
// through.  We keep going until we know that there aren't any lits
// left on the current level.  Then we stick the most recently
// resolved literal into the clause.

vector<literal> solver::combine_reasons(clause_id conf, fvector<char> &inclause)
{
  int lits_left = 0;
  literal p;
  vector<literal> ans(1);               // leave room for the asserting literal
  unsigned idx = stack.size() - 1;

  do {
    clause &c = *conf;
    if (c.learned) increase_value(c);

    for (auto l : c) 
      if (p != l && !inclause[l] && level[l]) {
        increase_value(atom(l));
        inclause[l] = true;
        if (level[l] >= curr_level()) ++lits_left; // current level; resolve away
        else ans.push_back(l);                     // earlier level; let stand
      }
        
    while (!inclause[stack[idx--]]);    // back up to inclause literal
    p = stack[idx + 1];                 // next literal (you went one too far)
    conf = reason[p];                   // next clause
    inclause[p] = false;                // clear inclause flag
    lits_left--;                        // one less to do
  } while (lits_left);
                                        // stop when all literals done
  ans[0] = -p;                          // establish reason for last literal
  return ans;
}

// Here is the simplification.  We want to remove all of the literals
// that aren't on an "interesting" level; in order to make the
// computation fast, we just record the "interestingness" mod 64, and
// figure that's good enough.  Here, we get the appropriate bit in a
// size_t.

size_t abstractify(unsigned level) 
{
  return (size_t) 1 << (level & (sizeof(size_t) * 8 - 1));
}

size_t solver::relevant_levels(const vector<literal> &cl)
{
  size_t ans = 0;
  for (auto l : cl) ans |= abstractify(level[l]);
  return ans;
}

// Simplify a clause.  Keep literals that are choices (i.e., have null
// reason) or are not otherwise removable.

vector<literal> solver::simplify(vector<literal> &cl, fvector<char> &inclause)
{
  vector<literal> toclear = cl;
  size_t abstract_level = relevant_levels(cl);
  max_lits += cl.size();        // all lits consider

  size_t j = 1;
  for (size_t i = 1; i < cl.size(); i++)
    if (!reason[cl[i]] || !removable(cl[i],abstract_level,toclear,inclause))
      cl[j++] = cl[i];
  cl.resize(j);
  tot_lits += cl.size();        // number that survived simplification

  for (auto l : toclear) inclause[l] = false; // reset inclause
  return cl;
}

// A literal can be removed if, as we amalgamate its reason and the
// reasons for the literals in said reason, we never get to an
// interesting level.

// So just go through them one at a time.  If at any point you find a
// literal that you haven't seen, and that isn't unit (ie., its level
// is positive), and that is assigned at an interesting level, you
// can't remove the original literal.  So clear what you need to clear
// and return false.

// Otherwise, keep going until there is nothing left to do, in which
// case you can return true.

bool solver::removable(literal p, size_t relevant_levels, 
                       vector<literal> &toclear, fvector<char> &inclause) 
{
  static vector<literal> possibles; // you call this a lot; set it up so that
  possibles.clear();                // you don't have to keep reinitializing
  possibles.push_back(p);           // the vector
  int top = toclear.size();
  while (!possibles.empty()) {
    const clause& c = *reason[possibles.back()];
    possibles.pop_back();

    for (size_t i = 1 ; i < c.size() ; i++) {
      literal l = c[i];
      if (!inclause[l] && level[l]) {
        if (reason[l] == clause_id() || 
            (abstractify(level[l]) & relevant_levels) == 0) {
          for (size_t j = top ; j < toclear.size() ; j++) 
            inclause[toclear[j]] = false;
          toclear.resize(top);
          return false;
        }
        else {
          inclause[l] = true;
          possibles.push_back(l);
          toclear.push_back(l);
        }
      }
    }
  }
  return true;
}

// Resolve a known conflict back to the bias.  Easy -- just keep
// resolving, pushing things that *aren't* in the bias on a list of
// literals still to do.

vector<literal> solver::resolve_to_bias(const vector<literal> &conf)
{
  static fvector<char> seen;     // faster than vector<bool>, I think
  if (seen.empty()) seen.resize(1 + size(),false);
  vector<literal> ans;
  vector<literal> todo = conf;
  vector<atom> done;
  while (!todo.empty()) {
    literal l = todo.back();
    todo.pop_back();
    if (seen[l] || !level[l]) continue;
    seen[l] = true;
    done.push_back(atom(l));
    if (!bias_true(l)) ans.push_back(l);
    else copy((*reason[l]).begin(),(*reason[l]).end(),back_inserter(todo));
    // gets l back on stack but it's already been seen
  }
  for (auto a : done) seen[a] = false;
  static fvector<char> inclause; // lits in clause
  if (inclause.empty()) inclause.resize(size(),false);
  for (auto l : ans) inclause[l] = true;
  return simplify(ans,inclause);       // also clears inclause
}

// PUBLIC METHOD: One search loop without restarts.  Stop after you've
// had the given number of conflicts (0 means never stop).  First
// increment the conflict limit; it's a delta and you've probably
// already got some.

// Now do the usual loop:
// 1. unit propagate
// 2. if a conflict is found, process it.  If the theory is UNSAT, you're 
//    done; otherwise, recur
// 3. if interrupting or no conflict and you've hit the search limit, return 
//    UNKNOWN
// 4. if simplification reveals a conflict, return UNSAT
// 5. otherwise, branch; if you can't branch, you found a model
// 6. make the branch literal true and continue.

Result solver::search(size_t conflict_limit)
{
  if (conflict_limit) conflict_limit += conflicts;

  bool pre_backtrack = true;
  for ( ;; ) {
    clause_id conf = unit_prop();                               // 1
    if (conf) {
      if (!new_conflict(conf)) return UNSAT;                    // 2
      pre_backtrack = false;
    }
    else if (interrupted || (conflict_limit && conflicts >= conflict_limit)) // 3
      return UNKNOWN; 
    else if (!simplify()) return UNSAT;                         // 4
    else {                      // no conflict and all ok
      if (learned.size() >= stack.size() + learned_limit) reduce(); // discard
      literal branch = pick_branch(pre_backtrack);              // 5
      if (branch == literal()) return SAT;                      // 5 - model
      if (data) note_heuristic(branch,data->heuristic_choice);
      new_decision();
      make_true(branch);                                        // 6 and recur
    }
  }
}

// Subsequences of the Luby-sequence:
//  0: 1
//  1: 1 1 2
//  2: 1 1 2 1 1 2 4
//  3: 1 1 2 1 1 2 4 1 1 2 1 1 2 4 8

size_t settings::luby(unsigned x) const
{
  if (compulsive) return 1;
  size_t subseq = 0, size = 1; // subseq is 0,1,2,3 as above; size is 1,3,7,15
  while (size <= x) {
    ++subseq;
    size = 2 * size + 1;
  } // now you have the subsequence

  while (size - 1 != x) {
    size = (size - 1)>>1;
    --subseq;
    x = x % size;
  } // now you've worked back to the actual exponent

  return restart_first * (size_t(1) << subseq);
}

// Top level solver.  Set the learned clause limit, then just search
// with the Luby numbers.  When you're done, copy the values into the
// model values for all the variables.  Clear the stack.

// If we're recording performance data, save it at the end of each
// probe.

// A few comments regarding diagnostics.  In the verbosity > 1 case,
// we show the current bias and polarity at each restart (the latter
// only if minisat is false).

Result solver::solve()
{
  learned_limit = clauses.size() / learned_factor;
  if (data) data->resize_to_nvars();
  Result ans = UNKNOWN;
  for (starts = 0 ; starts < probe_lim ; ++starts) {
    if (verbosity > 1) {
      cout << "PROBE " << starts << endl;
      if (!minisat) cout << "  bias " << bias << endl;
      cout << "  polarity " << polarity << endl;
    }
    bias_clause = clause_id();
    ans = search(luby(starts));
    if (interrupted) break;
    if (data) save_probe();
    update_variable_counts();
    if (ans == UNKNOWN && !minisat) ans = absorb_nogood()? UNKNOWN : UNSAT;
    if (ans != UNKNOWN) {
      ++starts;            // include solving probe, not just restarts
      break;
    }
    backtrack();
  }
  show_trailer();
  if (ans == SAT) {
    for (unsigned i = 0 ; i < size() ; ++i) model[i] = value[i];
    if (data) {
      data->convert_solution(*this);
      cout << *data;
    }
  }
  backtrack();
  return ans;
}
