// THIS SOURCE CODE IS SUPPLIED "AS IS" WITHOUT WARRANTY OF ANY KIND,
// AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH
// (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL
// WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND ANY
// WARRANTIES OR NON INFRINGEMENT.  THE USER ASSUMES ALL LIABILITY AND
// RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR
// NOR JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE
// FOR DAMAGES OF ANY KIND RESULTING FROM ITS USE.  Without limiting
// the generality of the foregoing, neither the author, nor JAIR, nor
// JAIR's publishers and distributors, warrant that the Source Code
// will be error-free, will operate without interruption, or will meet
// the needs of the user.

#ifndef __SOLVER__
#define __SOLVER__

#include "stl.h"
#include "pdb.h"
#include <cmath>
#include "performance.h"
#include <sys/time.h>
#include <sys/resource.h>

template <class T>
struct fvector : vector<T> {
  T      & operator[](unsigned n)      { return vector<T>::operator[](int(n)); }
  T      & operator[](atom a)          { return vector<T>::operator[](int(a)); }
  T      & operator[](literal l)       { return vector<T>::operator[](atom(l)); }
  const T& operator[](unsigned n) const { return vector<T>::operator[](int(n)); }
  const T& operator[](atom a)    const { return vector<T>::operator[](int(a)); }
  const T& operator[](literal l) const { return vector<T>::operator[](atom(l)); }
};

// Various settings controlling the search

struct settings {
  double        var_decay;              // v: variable decay factor
  double        clause_decay;           // c: clause decay factor
  double        garbage_frac;           // g: freed memory before gc
  int           restart_first;          // f: base restart interval
  bool          fastexit;               // x: just exit, don't return
  bool          print_model;            // m: show model if there is one
  bool          debug;                  // d: switch, settable at run time

  int           verbosity;              // V
  size_t        cpu_lim;                // C
  size_t        mem_lim;                // M
  size_t        probe_lim;              // l: max # of probes

  bool          minisat;                // s
  double        explore;                // T
  float         width;                  // w  [t -- total order -- is w=1,W=0]
  float         log_width;              // W
  bool          compulsive;             // b: full restart every node?

  bool          check_nogoods;          // n
  bool          performance;            // p (true means show it)
  bool          gzip;                   // z: input file is gzipped

  string        infile;

  // search control parameters (cannot be set on command line)
  double        learned_factor;         // initial learned limit = # clauses/this
  double        learned_mult;           // multiply by this much each time

  double        next_increment;         // increment things at this # conflicts
  double        display_delta;
  double        dd_factor;

  double        clause_delta;           // clause value increment
  double        variable_delta;         // variable value increment

  size_t        simplify_limit;         // simplify when this # propagations
  double        learned_limit;          // limit on # of learned clauses

  settings() : var_decay(0.95), clause_decay(0.999),
               garbage_frac(0.2), restart_first(100), fastexit(false),
               print_model(false), debug(false), verbosity(1),
               cpu_lim(-1), mem_lim(-1), probe_lim(-1),
               minisat(false), explore(0), width(1), log_width(0.1875),
               compulsive(false), check_nogoods(false), performance(false), 
               gzip(false),
               learned_factor(3), learned_mult(1.1),
               next_increment(100), dd_factor(1.5), 
               clause_delta(1), variable_delta(1),
               simplify_limit(0) {
    display_delta = dd_factor * next_increment; 
  }

  settings(size_t argc, char **argv);

  void set_cpu_limit() const;           // call operating system to set limits
  void set_memory_limit() const;        // based on cpu_lim or mem_lim

  size_t luby(unsigned x) const;
};

// Statistics about resources used by the search

class statistics {
  size_t starts;                // # restarts
  size_t decisions;             // # branch selections
  size_t propagations;          // # unit props
  size_t conflicts;             // # conflicts found
  size_t clause_lits;           // # lits original clauses
  size_t learned_lits;          // # lits learned clauses
  size_t max_lits;              // max # lits ever
  size_t tot_lits;              // total # lits (including deleted)
  double initial_time;          // time at start
  mutable problem_data *data;   // NULL unless you want to collect it
  bool   interrupted;           // have you been?

  statistics() : starts(0), decisions(0), propagations(0), conflicts(0),
                 clause_lits(0), learned_lits(0), max_lits(0), tot_lits(0),
                 initial_time(0), data(nullptr), interrupted(false) { }

  friend class solver;
  friend class problem_data;
  friend ostream &operator<<(ostream &os, const statistics &stats);
};

// State of the solver

struct state {
  // Note: these are not fvectors because the value is different for a
  // positive and negated literal.
  vector<vector<watcher>>   watches;    // lit -> places where it is a watcher
  vector<vector<clause_id>> goccurs;    // lit -> gamma clauses including it

  vector<literal>  stack;              // assignments, in order
  vector<unsigned> stack_sizes;        // sizes after each propagation
  unsigned         propagation_point;  // assignments from here forward need
                                       // to be propagated
  unsigned         assignment_limit;    // # of assignments at last reduction
  state() : propagation_point(0), assignment_limit(-1) { }
  friend ostream &operator<<(ostream &os, const state &s);
};

// solver class.  Settings, statistics and state, information
// associated with the variables being analyzed, clauses, and heap.
// (The heap has to be here because you need access to the VSIDS
// values.).  Also the Gamma set needed by PDB, and the best
// level/clause found so far so that you know what's "learned" from
// any particular probe.

class solver : public settings, public statistics, state {
  Dag                   dag;
  vector<clause_id>     clauses, learned;
  set<clause_id>        gamma;
  mg_heap               order_heap;  // variable heaped by value
  clause_id             bias_clause; // nogood for systematicity, this restart

public:
  solver(const settings &s) : settings(s), order_heap(this), heuristic_max(0) { }
  ~solver() { delete data; }

  void interrupt();
  void force_exit();

  void resize(unsigned n);
  unsigned size() const { return reason.size(); } // or whatever

private:
  // variable-associated data; stored this way to improve cache performance
  fvector<char>         known;     // value forced by (learned?) unit clause?
  fvector<clause_id>    reason;    // why does it have this value?
  fvector<unsigned>     level;     // level at which values were set
  fvector<literal>      value;     // value of branching on this variable
  fvector<double>       bvalue;    // value of branching on this variable
  fvector<literal>      polarity;  // preferred polarity
  fvector<literal>      bias;      // value in PDB bias
  fvector<literal>      model;     // value in model
  fvector<size_t>       count;
  double                heuristic_max; // maximum value at any given time

  // access functions for variable values
  bool      is_free    (atom a)    const { return value    [a].null(); }
  bool      is_false   (literal l) const { return value    [l] == -l; }
  bool      is_true    (literal l) const { return value    [l] == l; }
  bool      bias_true  (literal l) const { return bias     [l] == l; }
  bool      pol_true   (literal l) const { return polarity [l] == l; }
  bool      model_true (literal l) const { return model    [l] == l; }
  bool      not_false  (literal l) const { return value    [l] != -l; }

  vector<watcher>       &watch(literal l)       { return watches[int(l)]; }
  const vector<watcher> &watch(literal l) const { return watches[int(l)]; }

  vector<clause_id>       &goccur(atom    a)       { return goccur(literal(a)); }
  const vector<clause_id> &goccur(atom    a) const { return goccur(literal(a)); }
  vector<clause_id>       &goccur(literal l)       { return goccurs[int(l)]; }
  const vector<clause_id> &goccur(literal l) const { return goccurs[int(l)]; }

  // internal methods
  vector<literal> read_clause(const string &line);
  bool      add_clause(vector<literal> &c); 
                                // DESTRUCTIVE! return false if contradiction
  void      new_clause       (const vector<literal> &lits, bool learn);
  void      watch_clause     (const clause_id ci);
  void      unwatch_clause   (const clause_id ci);
  void      set_gamma        (clause_id c, literal conclusion);
  void      unset_gamma      (const clause_id c);
  void      unset_gamma      (vector<clause_id> &v)
    { for (auto &cid : v) unset_gamma(cid); }
  void      clear_gamma      ();
  void      set_bias         (literal l);
  void      set_polarity     (literal l);
  void      remove_clause    (const clause_id ci);
  void      remove_satisfied (vector<clause_id> &cs);
  void      reduce           ();        // remove half the clauses

  literal   pick_branch_pre_backtrack();
  literal   pick_branch_post_backtrack();
  literal   pick_branch(bool pre_backtrack);

  bool      ok_conclusion(literal l, const clause &c) const;
  bool      over_width(literal l, const clause &c) const;
  literal   find_conclusion(const clause &c);
  bool      bad_conclusion(literal l) const
    { return is_false(l) && known[l]; } // can't flip!
  unsigned  width_limit() const
    { return max(1,(int) (width + log_width * log2(starts))); }

  void      new_decision() { stack_sizes.push_back(stack.size()); }
  unsigned  curr_depth() const { return stack.size(); }
  unsigned  curr_level() const { return stack_sizes.size(); }
  void      make_true(literal p, const clause_id why = clause_id());
  clause_id unit_prop();        // return conflict if one found
  clause_id unit_prop(literal p);
  void      backtrack(unsigned level = 0);

  bool      absorb_nogood();
  // dag functions, operating on literals
  void      add_arc(atom a, atom b) { dag.add_arc(a,b); }
  void      weaken(atom a) { dag.weaken(a); }
  bool      equiv(literal l1, literal l2) const 
    { return dag.equiv(atom(l1),atom(l2)); }
  bool      incomparable(literal l1, literal l2) const
    { return dag.incomparable(atom(l1),atom(l2)); }
  bool      leq(literal l1, literal l2) const 
    { return dag.leq(atom(l1),atom(l2)); }

  void      simplify_gamma(literal l); 
  clause_id simp(clause_id cid);
  void      simplify_antecedent(literal l);
  void      make_parseable_literal(literal l);
  void      make_parseable(atom a);

  void      adjust_bias(const clause_id cl);
  bool      satisfy(int vals, const vector<int> &indices, const clause &cls)
    const;
  bool      breaks_gamma(int vals, const vector<clause_id> &clauses,
                         const VarSet &eq, const vector<int> &indices) const;
  bool      gamma_conflict(literal l, vector<int> *possibles = nullptr,
                           vector<atom> *new_indices = nullptr) const;
  void      set_best_possibility(const vector<int> &possibles, 
                                 const vector<atom> &equiv);
  void      update_variable_counts();

  double    get_uct_value(atom a) const;
  double    get_uct_value(int chosen, const vector<atom> &equiv) const;

  void      increase_value        (clause &c);
  void      increase_value        (atom a);
  void      decrease_var_delta    () { variable_delta *= 1 / var_decay; };
  void      decrease_clause_delta () { clause_delta *= 1 / clause_decay; }
  void      reheap();

  bool      satisfied       (const clause &c) const;
  bool      bias_satisfied  (const vector<literal> &c)  const;
  bool      model_satisfied (const vector<literal> &c)  const;
  bool      is_reason       (const clause_id ci) const
  { return is_true((*ci)[0]) && reason[(*ci)[0]] == ci; }
                                     // can't delete a clause if it's a reason
                                     // (and if so, it's a reason for first lit)
  bool      is_gamma       (const clause_id ci) const { return ci->gamma; }
  vector<literal> resolve        (clause_id conf); 
  vector<literal> resolve_to_bias (const vector<literal> &conf); 
  void            pdb_resolve    (literal l, set<clause_id> &done, 
                                  set<literal> &resolvent) const;
  vector<literal> pdb_resolve    (literal l, const clause_id conf); 
  vector<literal> combine_reasons(clause_id conf, fvector<char> &pending);
  size_t    relevant_levels(const vector<literal> &cl);
  vector<literal> simplify       (vector<literal> &cl, fvector<char> &inclause);
  template<class Lits>
  unsigned  get_level(const Lits &c) { return (c.size() > 1)? level[c[1]] : 0; }
  // first two atoms have to be deepest/2nd deepest for this to work!
  template<class Lits> void flevel (Lits &cl);
  void      find_level     (clause &cl);
  void      find_level     (vector<literal> &cl);
  bool      removable      (literal p, size_t abstract_levels,
                            vector<literal> &toclear, fvector<char> &pending);
                                // is p redundant?
  bool      new_conflict      (clause_id conf);
  clause_id new_learned_clause(vector<literal> &c);
  void      show_status();

  void     note_heuristic(literal l, vector<vector<size_t>> &bucketized) const;
  void     increment_heuristics(literal l) const;
  void     save_probe(const vector<literal> &learned_clause = vector<literal>())
    const;
  void     save_bias() const;

public:
  bool    read(istream &inf);
  bool    read_zip(const string &inf);

  bool    show_summary()         const;
  void    show_header (double t) const;
  void    show_trailer()         const;
  void    show_unit   ()         const;
  void    show_result (Result r) const;
  void    show_model  ()         const;

  void    check       (const string &desc) const; // debugging stub

  void    copy_model  (const solver &other);
  bool    simplify    ();       // remove satisfied clauses; return false if cnt
  Result  search      (size_t conflicts_limit); // one Luby loop
  bool    initialize();
  Result  solve(bool init_ok);
  Result  solve       ();

  // Memory managment:
  pool    litpool;
  virtual void garbage_collect();
  void    check_garbage() { 
    if (litpool.free > litpool.size() * garbage_frac) garbage_collect(); 
  }
  void    reloc_all(pool& to);
  void    add_to_gamma(clause_id cid);
  void    recompute_gamma();

  friend class problem_data;
  friend bool mg_heap::cmp(atom x, atom y) const;
  friend ostream &operator<<(ostream &os, const solver &s) 
  { return os << state(s); }
};

inline clause       & clause_id::operator*() { return (clause &) (*p)[index]; }
inline const clause & clause_id::operator*() const {
  return (const clause &) (*p)[index]; 
}

double get_cpu_time();
double mem_used();
void set_limit(int resource, rlim_t limit, const string &desc);

#endif
