// THIS SOURCE CODE IS SUPPLIED "AS IS" WITHOUT WARRANTY OF ANY KIND,
// AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH
// (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL
// WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND ANY
// WARRANTIES OR NON INFRINGEMENT.  THE USER ASSUMES ALL LIABILITY AND
// RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR
// NOR JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE
// FOR DAMAGES OF ANY KIND RESULTING FROM ITS USE.  Without limiting
// the generality of the foregoing, neither the author, nor JAIR, nor
// JAIR's publishers and distributors, warrant that the Source Code
// will be error-free, will operate without interruption, or will meet
// the needs of the user.

#ifndef __STL__
#define __STL__

#include <cassert>
#include <list>
#include <vector>
#include <set>
#include <iostream>
#include <numeric>
#include <algorithm>

using namespace std;

enum Result { UNSAT, SAT, UNKNOWN };

struct literal;

struct atom {
  unsigned                x;                      
  atom(unsigned n = -1) : x(n) { }
  atom(int n)           : x(n) { }
  atom(size_t n)        : x(n) { }
  explicit atom(const literal l); // atom from literal ok but you have to ask

  template<class T> atom(T t) = delete;
  operator unsigned() const   { return x; }

  literal operator-() const;    // produces negated literal
  literal assigned_val(size_t idx, int vals) const;

  friend ostream &operator<<(ostream &os, atom a);
};

struct literal {
  unsigned     x;               // low order bit 1 means negated

  literal(unsigned n = -1)          : x(n) { }
  literal(int n)                    : x(n) { }
  literal(size_t n)                 : x(n) { }
  literal(atom a, bool neg = false) : x(a << 1 | neg) { }
  template<class T> literal(T t) = delete;

  bool null()                           const { return x == (unsigned) -1; }
  bool negated()                        const { return x & 1; }
  bool operator==(const literal &other) const { return x == other.x; }
  bool operator!=(const literal &other) const { return x != other.x; }
  bool operator< (const literal &other) const { return x < other.x; }
  literal operator-()                   const { return x ^ 1; }  // negation

  explicit operator int()               const { return x; }
  int to_int() const {
    unsigned a(1 + atom(*this));
    return negated()? -a : a;
  }

  friend ostream &operator<<(ostream &os, literal l);
};

inline literal atom::operator-() const { return literal(*this,true); }
inline atom::atom(const literal l)     : x(l.x >> 1) { }
inline literal int_to_lit(int x)       { return literal(abs(x) - 1,x < 0); }
                                // for reading cnf, where -x is negated

// A clause is a block of literals in the literal pool, together with
// a bool indicating whether it's learned (or original) and the VSIDS
// value computed from recent activity.  Also PDB stuff: a flag indicating
// that the clause is in Gamma and shouldn't be deleted, and an
// indication of which literal is the conclusion.

struct clause_header  {
  unsigned learned       : 1;
  mutable unsigned gamma : 1;   // is clause in Gamma?
  unsigned hsize         : 30; 
  float    value;
  mutable literal conclusion;   // literal in equivalence class of conclusion

  clause_header() : learned(false), hsize(0), value(0) { }
  clause_header(bool l, unsigned s) : learned(l), hsize(s), value(0) { }
};

struct clause_id;

struct clause : clause_header {
  literal   data[0];            // uses no space; it's all in the literal pool

  // vector-like functionality, to the extent possible (can't grow a clause)
  unsigned      size        ()      const   { return hsize; }
  bool          empty       ()      const   { return size() == 0; }
  void          resize      (unsigned i)    { 
    assert(i <= size());
    hsize = i;
  }
  void          pop_back    ()              { resize(size() - 1); }
  const literal &back       ()      const   { return data[hsize - 1]; }

  literal & operator []     (int i)         { return data[i]; }
  literal   operator []     (int i) const   { return data[i]; }

  literal       * begin()       { return &data[0]; }
  literal       * end  ()       { return begin() + size(); }
  const literal * begin() const { return &data[0]; }
  const literal * end  () const { return begin() + size(); }

  bool   properly_subsumes(const clause    &other) const;
  bool   properly_subsumes(const clause_id &other) const;

  friend ostream &operator<<(ostream &os, const clause &cl);
};

// A clause_id locates a clause in the literal pool.  Note that it is
// an unsigned instead of a size_t; the literal pool is assumed <3G in
// size (each element itself being an unsigned), and unsigned are
// smaller.

class pool;

struct clause_id {
  unsigned     index;
  static pool  *p;       // for dereferencing

  clause_id(size_t i = -1) : index(i) { }

  bool operator==(const clause_id other) const { return index == other.index; }
  bool operator!=(const clause_id other) const { return index != other.index; }
  bool operator< (const clause_id other) const { return index <  other.index; }

  operator bool() const { return *this != clause_id(); }

  clause       &operator*();
  const clause &operator*() const;
  clause       *operator->()       { return &operator*(); }
  const clause *operator->() const { return &operator*(); }
};

// This describes how the literal pool is set up.  

// First, we want to have a literal pool; saves about 20% with one
// thread and about 50% with 8 threads.

// The problem is that clauses have ancillary information beyond
// what's in the literal pool -- size and value, and perhaps more.
// The value isn't really "part" of the clause; that should arguably
// be stored off to the side in some way and the literal pool should
// just include the size and the actual literals.

// The problem is that is that you need pointers to clauses all over.
// So, for example, you could have a pointer to the spot in the
// literal pool where the clause information is, together with a
// pointer to the location of the ancillary information.  This second
// pointer is a lot of extra bookkeeping, though; it's simpler to just
// drop the ancillary information into the literal pool.

// So the literal pool itself consists of a bunch of blocks, each of
// which is a header and then the literals in question.  We allocate
// the header separately, but the pool is basically just a vector of
// lits.

// We also store a bit locally indicating whether or not this clause
// is learned.  (Doing this lets us avoid bumping the value of the
// clause if it's not learned, which doesn't really matter so much
// from a timing perspective but *does* help to avoid reducing the
// values so often.)  We just crowd that bit into the unsigned giving
// the length of the clause.

class pool : public vector<literal>
{
  static int header_size() {
    return (sizeof(clause_header) + sizeof(literal) - 1) / sizeof(literal); }
  // just in case header is not an integral number of literals in size

public:
  unsigned free;
  pool(size_t sz = 0) : vector<literal>(sz), free(0) {}

  // Push a new clause into the literal pool.  Resize it to make room
  // and then just treat the new memory as a clause.  Write out all
  // the relvant data and return a clause_id.
  template<class literals>
  clause_id alloc(const literals& ps, bool learned = false) {
    unsigned sz = size();
    resize(size() + header_size() + ps.size());
    clause &cl = (clause &) (*this)[sz];
    cl.learned = learned;
    cl.gamma   = false;
    cl.hsize   = ps.size();
    cl.value   = 0;
    for (size_t i = 0 ; i < ps.size() ; ++i) cl.data[i] = ps[i];
    return clause_id(sz);
  }

  // To "release" a clause, just increase the amount of
  // known-to-be-free space.  Garbage collection will eventually
  // release it.
  void release(clause_id sz) { free += header_size() + sz->size(); }

  void reloc(clause_id &cr, vector<clause_id> &relocs, pool& to);
};

// A watcher indicates a place where a variable is watched.  That
// means that it's basically a clause_id, but also includes the other
// watched literal.  In the event that we update the watcher and need
// to find a new watcher for the clause in question, we check the
// other watched literal first.

struct watcher : clause_id {
  literal  other;
  watcher() {}
  watcher(clause_id c, literal o) : clause_id(c), other(o) {}
};

class solver;

// Priority queue where priority can change on the fly

class mg_heap : public vector<atom> {
  vector<int>      location; // where is this atom in the heap? (-1 if unplaced)
  const solver     *slv;     // pointer; needed for comparison function

 public:
  bool cmp(atom i, atom j) const;
  void move_up(int idx);
  void move_down(int idx);

  mg_heap(const solver *s) : slv(s) { }

  void clear();
  void insert(atom v);
  void update(atom v) { if (contains(v)) move_up(location[v]); }
                                // assumes score got bigger!
  atom front();                 // pops off queue as well
  bool contains(atom v) const { return v < location.size() && location[v] >= 0; }
  void insert_if_missing(atom v) { if (!contains(v)) insert(v); }
  void rebuild(const vector<atom> &v);
};

// display a vector or set by displaying the elements
template<class T>
static ostream &operator<<(ostream &os, const vector<T> &v)
{
  for (auto &i : v) os << ((&i == &*v.begin()) ? "" : " ") << i;
  return os;
}

template<class T>
static ostream &operator<<(ostream &os, const set<T> &s)
{
  for (auto &i : s) os << ((&i == &*s.begin()) ? "" : " ") << i;
  return os;
}

#endif
