// THIS SOURCE CODE IS SUPPLIED "AS IS" WITHOUT WARRANTY OF ANY KIND,
// AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH
// (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL
// WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED WARRANTIES OF
// MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND ANY
// WARRANTIES OR NON INFRINGEMENT.  THE USER ASSUMES ALL LIABILITY AND
// RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR
// NOR JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE
// FOR DAMAGES OF ANY KIND RESULTING FROM ITS USE.  Without limiting
// the generality of the foregoing, neither the author, nor JAIR, nor
// JAIR's publishers and distributors, warrant that the Source Code
// will be error-free, will operate without interruption, or will meet
// the needs of the user.

#include "solver.h"
#include <iomanip>

// Satisfied is easy.

bool solver::satisfied(const clause &c) const 
{
  return any_of(c.begin(),c.end(),[&](literal l) { return is_true(l); });
}

// Satisfied in the bias is the same, basically

bool solver::bias_satisfied(const vector<literal> &c) const 
{
  return any_of(c.begin(),c.end(),[&](literal l) { return bias_true(l); });
}

// So is satisfied in the known model

bool solver::model_satisfied(const vector<literal> &c) const 
{
  return any_of(c.begin(),c.end(),[&](literal l) { return model_true(l); });
}

// Level of a clause.  0 if unit, otherwise swap the most recently
// assigned literal to position 1 and return its level.

template<class Lits>
void solver::flevel(Lits &cl)
{
  if (cl.size() == 1) return;

  unsigned max_lev = 0, second_lev = 1;
  if (level[cl[0]] < level[cl[1]]) { max_lev = 1; second_lev = 0; }

  for (unsigned i = 2 ; i < cl.size() ; ++i)
    if (level[cl[i]] > level[cl[max_lev]]) { second_lev = max_lev; max_lev = i; }
    else if (level[cl[i]] > level[cl[second_lev]]) second_lev = i;

  swap(cl[0],cl[max_lev]);                    // first element highest level
  if (second_lev) swap(cl[1],cl[second_lev]); // second element next highest
  else swap(cl[max_lev],cl[1]);               // second element was first; get it
}

void solver::find_level(clause &cl)          { flevel(cl); }
void solver::find_level(vector<literal> &cl) { flevel(cl); }

// Various display functions

// Print a literal as a signed atom and a clause (a block in the
// literal pool) as a sequence of literals.

ostream &operator<<(ostream &os, literal l) { return os << l.to_int(); }
ostream &operator<<(ostream &os, const atom a) { return os << literal(a); }

ostream &operator<<(ostream &os, const clause &cl) {
  for (size_t i = 0 ; i < cl.size() ; ++i) os << (i? " " : "") << cl[i];
  return os;
}

void solver::show_status() 
{
  cout << "|" << setw(10) << conflicts << " |" << setw(8)
       << size() - (stack_sizes.empty() ? stack.size() : stack_sizes[0])
       << setw(9) << clauses.size() << setw(9) << clause_lits << " |"
       << setw(9) << int(learned_limit) << setw(9) << learned.size()
       << setw(7) << int(double(learned_lits) / learned.size() + 0.5) 
       << "  |" << endl;
}

ostream &operator<<(ostream &os, const statistics &stats)
{
  double cpu_time = get_cpu_time() - stats.initial_time;
  double mem = mem_used();
  int p = os.precision();

  os << "restarts              : " << stats.starts << endl
     << "conflicts             : " 
     << left << setw(15) << stats.conflicts << " (" << fixed << setprecision(0) 
     << stats.conflicts / cpu_time << "/sec)" << endl
     << "decisions             : " << left << setw(16) << stats.decisions
     << "(" << setprecision(0) << stats.decisions / cpu_time << "/sec)" << endl
     << "propagations          : " 
     << left << setw(15) << stats.propagations << " ("
     << stats.propagations / cpu_time << "/sec)" << endl
     << "conflict literals     : " 
     << left << setw(15) << stats.tot_lits << " (" << setprecision(2) 
     << 100.0 * (stats.max_lits - stats.tot_lits) / stats.max_lits
     << "% deleted)" << endl;
  if (mem)
    os << "Memory used           : " << setprecision(2) << mem << " MB"
       << endl;
  return os << setprecision(p) << "CPU time              : " << cpu_time
            << " s" << endl;
}

// Display a state; displaying a solver defaults to this.  Just print
// out the current assignment stack but mark choices for easy
// identification.

ostream &operator<<(ostream &os, const state &s)
{
  for (size_t i = 0 ; i < s.stack.size() ; ++i) {
    if (i) os << ' ';
    if (find(s.stack_sizes.begin(),s.stack_sizes.end(),i) != 
        s.stack_sizes.end())
      os << "*";
    os << s.stack[i];
  }
  return os;
}

// Display a dag.  Just show the skeleton of <=.

ostream &operator<<(ostream &os, const Dag &d)
{
  for (size_t i = 0 ; i < d.size() ; ++i) {
    if (d.followers(i).size() == 1) continue;
    const VarSet &e = d.equiv_class(i);
    os << i;
    if (e.size() > 1) {
      os << " =";
      for (size_t j = 0 ; j < e.size() ; ++j) if (e[j] != i) os << ' ' << e[j];
    }
    const VarSet &f = d.proper_followers(i);
    if (!f.empty()) os << " < " << f;
    os << endl;
  }
  return os;
}

